#include "../stdAfx.h"
#include "resource.h"
#include "out.h"
#include <math.h>
#include <assert.h>
#include <mmreg.h>
#include "AUDIO.H"
#include "dllOutput.h"
#include "../wa_ipc.h"
#include "../MainHelper.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#ifndef _WAVEFORMATEXTENSIBLE_
#define _WAVEFORMATEXTENSIBLE_
typedef struct
{
	WAVEFORMATEX Format;
	union
	{
		WORD wValidBitsPerSample;   	/* bits of precision  */
		WORD wSamplesPerBlock;  		/* valid if wBitsPerSample==0 */
		WORD wReserved; 				/* If neither applies, set to zero. */
	} Samples;
	DWORD dwChannelMask;	  /* which channels are */
	/* present in stream  */
	GUID SubFormat;
} WAVEFORMATEXTENSIBLE, * PWAVEFORMATEXTENSIBLE;
#endif // !_WAVEFORMATEXTENSIBLE_

#ifndef DWORD_PTR
typedef ULONG DWORD_PTR;
typedef long LONG_PTR;
#endif

#if !defined(WAVE_FORMAT_EXTENSIBLE)
#define  WAVE_FORMAT_EXTENSIBLE 				0xFFFE /* Microsoft */
#endif // !defined(WAVE_FORMAT_EXTENSIBLE)

TCHAR outDescription[50] = _T("Module standard redvelopper par Mickal SEBIRE");

Helper m_HelperOut;

int genVolume = 0;

/*
 * private "members"
 */
static CRITICAL_SECTION critical_section;
static HWAVEOUT device;
static WAVEHDR* headers;
static int current_header;
static int free_headers;
static int written;
static int output;
static int pos;
static int volume;
static int pan;
static int bytes_per_sec;
static int paused;
static config_t config_master;
static config_t config_current;
static format_t format;
static prefsDlgRec outConfig;

/*
 * get name of configuration file
 */
void config_get_filename(TCHAR* buffer, int size)
{
	TCHAR* p;

	assert(buffer);

	memset(buffer, 0, size);

	GetModuleFileName(NULL, buffer, size);
	p = buffer + _tcslen(buffer);

	while (p >= buffer && *p != '.')
		p--;

	_tcscpy(p, _T(".ini"));
}


/*
 * load configuration from a file and fill config with it
 */
int config_load(config_t* config)
{
	TCHAR buffer[MAX_PATH];

	assert(config);

	config_get_filename(buffer, sizeof(buffer));

	return GetPrivateProfileStruct(_T("out_ivwave"), _T("config"), config, sizeof(config_t), buffer);
}

/*
 * save configuration to a file from config.
 */
void config_save(const config_t* config)
{
	TCHAR buffer[MAX_PATH];

	assert(config);

	config_get_filename(buffer, sizeof(buffer));
	WritePrivateProfileStruct(_T("out_ivwave"), _T("config"), (void*)config, sizeof(config_t), buffer);
}

/*
 * fill config with the default settings
 */
void config_set_defaults(config_t* config)
{
	assert(config);
	config->block_count = BLOCK_COUNT;
	config->block_size = BLOCK_SIZE;
	config->device = -1;
	config->reverse_balance = 0;
	config->volume_enabled = 1;
	config->pcmconv_enabled = 1;
}

/*
 * init - called once when the plug-in is loaded. this simply
 * initialises some static vars which are already initialised (if my
 * understanding of the C standard is correct).
 */
void OutputInit()
{
	device = NULL;
	headers = NULL;
	volume = ::SendMessage(standardOut.hMainWindow, WM_WA_IPC, -666, IPC_SETVOLUME);
	pan = 0;
	static CString name = m_HelperOut.MyLoadString(IDD_LCONFIGDLG);

	outConfig._id = 5;
	outConfig.dlgID = IDD_CONFIGDLG;
	outConfig.name = (TCHAR *)(LPCTSTR)name;
	outConfig.proc = config_dialog_proc;
	outConfig.hInst = AfxGetInstanceHandle();
	outConfig.where = 4;
	::SendMessage(standardOut.hMainWindow, WM_WA_IPC, (WPARAM)&outConfig, IPC_ADD_PREFS_DLG);

	if (!config_load(&config_master))
	{
		/*
													 * defaults
													 */
		config_set_defaults(&config_master);
	}
	else
	{
		/* 
													 * validate configuration
													 */
		if (config_master.block_count < BLOCK_COUNT_MIN)
			config_master.block_count = BLOCK_COUNT_MIN;
		else if (config_master.block_count > BLOCK_COUNT_MAX)
			config_master.block_count = BLOCK_COUNT_MAX;

		if (config_master.block_size < BLOCK_SIZE_MIN)
			config_master.block_size = BLOCK_SIZE_MIN;
		else if (config_master.block_size > BLOCK_SIZE_MAX)
			config_master.block_size = BLOCK_SIZE_MAX;

		if (config_master.device > (int)waveOutGetNumDevs() - 1)
			config_master.device = -1;
		else if (config_master.device < -1)
			config_master.device = -1;
	}

	/*
	 * set up some bits to make the config display properly
	 * before we've initiated playback. since the rest are static
	 * we don't need to touch them.
	 */
	free_headers = config_master.block_count;
	config_current = config_master;
}

/*
 * open - open the device. srate = sampling rate, nch = number of
 * channels, bps = number of bits per sample. buflen and prebuf are 
 * ignored.
 */
int OutputOpen(int nSamplerate, int nNumchannels, int nBitspersamp, int bufferlenms, int prebufferms)
{
	//Nb octet = nSamplerate * bitpersamp
	//705600 = 44100 * 16
	//Time = 
	//TRACE(_T("OutputOpen %d %d %d %d %d\n"), nSamplerate, nNumchannels, nBitspersamp, bufferlenms, prebufferms);

	WAVEFORMATEX wfx;
	MMRESULT result;
	unsigned char* buffer;
	int i;

	assert(device == NULL);

	/*
	 * copy the config so that we can modify the original during
	 * playback without screwing anything up.
	 */
	config_current = config_master;

	format.bitspersamp = nBitspersamp;
	format.numchannels = nNumchannels;
	format.samplerate = nSamplerate;
	format.conv = 0;

	if (waveOutGetNumDevs() == 0)
	{
		/*  	  MessageBox(
																								ivwave.hMainWindow,
																								"Playback cannot proceed because there are no valid wave devices\n"
																								"present on this system. If you have a sound card and you are\n"
																								"sure that the drivers are correctly installed, consider trying\n"
																								"the DirectSound output plug-in instead.",
																								NAME" Error",
																								MB_ICONERROR
																							);
																					*/  	  return -1;
	}

	/*
	 * set up format info stuff
	 */
	wfx.cbSize = 0;
	wfx.wFormatTag = WAVE_FORMAT_PCM;
	wfx.nChannels = nNumchannels;
	wfx.nSamplesPerSec = nSamplerate;
	wfx.wBitsPerSample = nBitspersamp;
	wfx.nBlockAlign = (wfx.nChannels * wfx.wBitsPerSample) >> 3;
	wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;

	/*
	 * open the wave output device
	 */
	if ((result = waveOutOpen(&device, config_current.device, &wfx, (DWORD_PTR)waveout_callback, (DWORD_PTR)&free_headers, CALLBACK_FUNCTION)) !=
		MMSYSERR_NOERROR && config_current.pcmconv_enabled)
	{
		/*
																									 * see http://www.microsoft.com/hwdev/tech/audio/multichaud.asp
																									 *
																									 * As well as supporting audio with > 2 channels it will
																									 * perform conversions on the data before writing it to the
																									 * card.
																									 */
		WAVEFORMATEXTENSIBLE wfex;

		/*
																									 * avoiding ks.h and ksmedia.h
																									 */ 
		GUID KSDATAFORMAT_SUBTYPE_PCM =
		{
			0x00000001, 0x0000, 0x0010, { 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 }
		};

		wfex.Format = wfx;
		wfex.Format.cbSize = 22;
		wfex.Format.wFormatTag = WAVE_FORMAT_EXTENSIBLE;
		wfex.Samples.wReserved = 0;
		wfex.dwChannelMask = 0;
		wfex.SubFormat = KSDATAFORMAT_SUBTYPE_PCM;

		result = waveOutOpen(&device, config_current.device, (WAVEFORMATEX *)&wfex, (DWORD_PTR)waveout_callback, (DWORD_PTR)&free_headers,
							 CALLBACK_FUNCTION);

		format.conv = 1;
	}

	if (result != MMSYSERR_NOERROR)
	{
		WAVEOUTCAPS woc;
		TCHAR buffer[512];

		if (waveOutGetDevCaps((UINT)config_current.device, &woc, sizeof(WAVEOUTCAPS)) != MMSYSERR_NOERROR)
			_tcscpy(woc.szPname, _T("[Invalid Device]"));

		/*
																									 * argh
																									 */
		switch (result)
		{
		case WAVERR_BADFORMAT:
			_tprintf(buffer, "The \"%s\" device returned a Bad Format error (%d) during\n"
							 "the waveOutOpen call. This is because the device does not support the\n" "following PCM format:\n\n"
							 "Sample Rate:\t%d Hz\n" "Sample Size:\t%d bits (%s)\n" "Channels:\t%d\n\n"
							 "%sinput plug-in has options for changing the output format\n"
							 "try adjusting these to a format supported by your sound card.\n"
							 "Consult your sound card manual for exact format specifications.", woc.szPname, result, wfx.nSamplesPerSec,
					 wfx.wBitsPerSample, wfx.wBitsPerSample == 8 ? "unsigned" : "signed", wfx.nChannels,
					 config_current.pcmconv_enabled ? "If the current " :
					 "You might be able to fix this problem by enabling the \"PCM Conversion\"\n"
					 "option in the configuration for this plug-in. Alternatively, if the\ncurrent ");
			break;
		case MMSYSERR_BADDEVICEID:
			_tprintf(buffer, "The waveOut system returned a Bad Device ID error (%d) during\n"
							 "the waveOutOpen call. The device identifier given was %d.\n"
							 "Please ensure that you have a soundcard and that the drivers are\n"
							 "installed correctly. If this is already the case, try selecting a\n"
							 "different device from the \"Output Device\" section in the configuration\n"
							 "for this plug-in\n", result, config_current.device);
			break;
		default:
			_tprintf(buffer, "An MMSYSTEM error occurred while opening the wave device.\n\n" "Output device: %s\n"
							 "Error code: %u\n\n", woc.szPname, result);
		}

		/*  	  MessageBox(
																												ivwave.hMainWindow,
																												buffer,
																												NAME" Error",
																												MB_ICONERROR
																											);
																									*/  	  return -1;
	}

	/*
	 * allocate memory for buffers, headers etc.
	 */
	if ((buffer =
		 (unsigned char*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, (sizeof(WAVEHDR) + config_current.block_size) * config_current.block_count)) ==
		NULL)
	{
		waveOutClose(device);
		device = NULL;
		/*  	  MessageBox(
																			ivwave.hMainWindow,
																			"Memory allocation error allocating wave buffer",
																			NAME" Error",
																			MB_ICONERROR
																		);
																*/  	  return -1;
	}

	InitializeCriticalSection(&critical_section);

	headers = (WAVEHDR *)buffer;
	current_header = 0;
	free_headers = config_current.block_count;
	written = 0;
	output = 0;
	pos = 0;
	bytes_per_sec = nNumchannels * (nBitspersamp >> 3) * nSamplerate;
	paused = 0;

	/*
	 * set up buffer pointers
	 */
	for (i = 0; i < config_current.block_count; i++)
	{
		headers[i].lpData = (CHAR *)buffer + (sizeof(WAVEHDR) * config_current.block_count) + (i * config_current.block_size);
		headers[i].dwBufferLength = config_current.block_size;
	}

	/*
	 * now actually set volume and pan in case they've been
	 * set before the device was open
	 */
	OutputSetPan(pan);
	OutputSetVolume(volume);

	/*
	 * return max amount of time it'll take to play through the buffer in ms
	 * the +1 is a safety margin which adds approximately the time to play 1 
	 * block to the total.
	 */
	return MulDiv((config_current.block_count + 1) * config_current.block_size, 1000, bytes_per_sec);
}
/*
 * close - close the device if it is open. calling this with
 * the device closed will trigger an assert failure in debug mode.
 */
void OutputClose()
{
	//TRACE(_T("OutputClose\n"));
	int i;

	assert(device != NULL);

	/*
	 * unpause if paused
	 */
	if (paused)
	{
		waveOutRestart(device);
		paused = 0;
	}

	/*
	 * reset the device
	 */
	waveOutReset(device);

	/*
	 * wait for all buffers to complete
	 */
	while (free_headers != config_current.block_count)
		Sleep(10);

	/* 
	 * unprepare any prepared headers
	 */
	for (i = 0; i < config_current.block_count; i++)
	{
		if (headers[i].dwFlags & WHDR_PREPARED)
			waveOutUnprepareHeader(device, &headers[i], sizeof(WAVEHDR));
	}

	DeleteCriticalSection(&critical_section);

	/*
	 * close the device
	 */
	waveOutClose(device);
	device = NULL;

	/*
	 * and finally free the used memory
	 */
	HeapFree(GetProcessHeap(), 0, headers);
	headers = NULL;
}

/*
 * quit - called on quit. this assert checks that the device
 * was properly closed before the application unloaded the plug-in.
 * does nothing in release mode.
 */
void OutputQuit()
{
	//TRACE(_T("OutputQuit\n"));
	assert(device == NULL);
	::SendMessage(standardOut.hMainWindow, WM_WA_IPC, (WPARAM)&outConfig, IPC_REMOVE_PREFS_DLG);
}

/*
 * pause - pause or resume playback
 */
int OutputPause(int pause)
{
	//TRACE(_T("OutputPause %d\n"), pause);
	assert(device);

	if (pause)
	{
		waveOutPause(device);
		paused = 1;
	}
	else
	{
		waveOutRestart(device);
		paused = 0;
	}

	return !pause;
}

/*
 * write - this will actually block if you write too much
 * but the input plug-in should be smart enough to realise that it
 * should write 
 * a) no more than 8192 bytes and
 * b) no more than can_write() returns. 
 * we could add another size check here but it's not really worth it.
 */
int OutputPlay(char* buffer, int size)
{
	WAVEHDR* current;
	int remain;
	char* buf = (char*)buffer;

	assert(device);
	assert(size <= 8192);

	current = &headers[current_header];

	assert(current);

	written += size;

	while (size > 0)
	{
		OutputSetVolume(genVolume);
		/* 
																 * first make sure the header we're going to use is unprepared
																 */
		if (current->dwFlags & WHDR_PREPARED)
			waveOutUnprepareHeader(device, current, sizeof(WAVEHDR));

		if (size < (int)(config_current.block_size - current->dwUser))
		{
			memcpy(current->lpData + current->dwUser, buf, size);
			current->dwUser += size;

			break;
		}

		remain = config_current.block_size - current->dwUser;
		memcpy(current->lpData + current->dwUser, buf, remain);
		size -= remain;
		buf += remain;
		current->dwBufferLength = config_current.block_size;

		//written += config_current.block_size;

		waveOutPrepareHeader(device, current, sizeof(WAVEHDR));
		waveOutWrite(device, current, sizeof(WAVEHDR));

		/*
																 * writing to free_headers must be encased in 
																 * Enter/LeaveCriticalSection calls since two different
																 * threads write to it (reading should be ok though).
																 */
		try
		{
			EnterCriticalSection(&critical_section);
			free_headers--;
			LeaveCriticalSection(&critical_section);

			while (!free_headers)
				Sleep(10);

			current_header++;
			current_header %= config_current.block_count;

			current = &headers[current_header];
			current->dwUser = 0;
		}
		catch (...)
		{
		}
	}

	return 0;
} 

/*
 * flush - flush output buffers and continue playback from
 * 'time'.
 */
void OutputDrain(int time)
{
	//TRACE(_T("OutputDrain %d\n"), time);
	WAVEHDR* current;
	assert(device);

	current = &headers[current_header];

	/*
	 * if the current buffer has some data in it then write it
	 * anyway.
	 */
	if (current->dwUser)
	{
		current->dwBufferLength = current->dwUser;
		current->dwUser = 0;

		waveOutPrepareHeader(device, current, sizeof(WAVEHDR));
		waveOutWrite(device, current, sizeof(WAVEHDR));

		EnterCriticalSection(&critical_section);
		free_headers--;
		LeaveCriticalSection(&critical_section);
	}

	/*
	 * reset the device - take note that the microsoft documentation
	 * states that calling other waveOut functions from the callback
	 * will cause a deadlock - well calling waveOutUnprepare header
	 * doesn't until you call waveOutReset, at which point the app
	 * will freeze (hence the lame arse implementation here).
	 */
	waveOutReset(device);

	/*
	 * wait for all buffers to be returned by the reset
	 */ 
	while (free_headers != config_current.block_count)
		Sleep(10);

	/*
	 * point to the next block
	 */
	current_header++;
	current_header %= config_current.block_count;
	/*
	 * reset positioning/statistics variables
	 */
	output = 0;
	written = 0;
	pos = time;

	/*
	 * playback can now continue
	 */
}

/*
 * set_volume - set the volume (volume ranges from 0->255)
 */
void OutputSetVolume(int volume)
{
	//TRACE(_T("OutputSetVolume %d\n"), volume);
	genVolume = volume;
	unsigned short left;
	unsigned short right;

	/*
	 * out of range so do the default (use what the system is set to)
	 */
	if (volume < 0 || volume > 255)
		return;

	volume = volume;

	/*
	 * set_volume can be called before the device
	 * is open for defaults so check now.
	 */ 
	if (!device)
		return;

	if (config_current.volume_enabled)
	{
		volume = volume * 0xffff / 0xff;
		left = (unsigned short) volume;
		right = (unsigned short) volume;

		/*
																			 * compensate for the pan setting
																			 */
		if (pan > 0)
			left -= pan * volume / 128;
		else if (pan < 0)
			right -= -pan * volume / 128;

		/* 
																			 * and actually set the volume
																			 */
		if (config_current.reverse_balance)
			waveOutSetVolume(device, (right & 0xffff) | ((left & 0xffff) << 16));
		else
			waveOutSetVolume(device, (left & 0xffff) | ((right & 0xffff) << 16));
	}
}

/*
 * config - called by the parent application to display
 * a configuration dialog box.
 */
void OutputConfig(HWND hwndParent)
{
	::SendMessage(standardOut.hMainWindow, WM_WA_IPC, (WPARAM)&outConfig, IPC_OPENPREFSTOPAGE);
}

/*
 * about - called by the parent application to display
 * some information about the plug-in.
 */
void OutputAbout(HWND hwndParent)
{
	//TRACE(_T("OutputAbout\n"));
	CString message(outDescription);
	AfxMessageBox(message, MB_ICONINFORMATION | MB_OK);
}

/*
 * can_write - return number of bytes that can be written - 
 * 0 if device is closed - note that this can be called before playback
 * has been started.
 */
int CanWrite()
{
	//TRACE(_T("CanWrite\n"));

	WAVEHDR* current = &headers[current_header];

	if (!device)
		return 0;

	return free_headers * config_current.block_size - current->dwUser;
}

/*
 * playing - returns 1 if blocks are currently submitted
 * to the wave functions.
 */
int IsPlaying()
{
	//TRACE(_T("IsPlaying\n"));
	if (!device)
		return 0;

	return (free_headers != config_current.block_count);
}

/*
 * set_pan - set the pan (pan ranges from -128 to +128)
 */
void OutputSetPan(int pan)
{
	//TRACE(_T("OutputSetPan\n"));
	/*
	 * out of range so do the default (centre)
	 */
	if (pan < -128 || pan > 128)
	{
		pan = 0;
		return;
	}

	pan = pan;

	/* 
	 * set_pan can be called before the
	 * device is opened for setting the defaults
	 */
	if (!device)
		return;

	/*
	 * windows waveOut uses waveOutSetVolume to set l&r
	 * volumes together so leave the volume setter to deal
	 * with it.
	 */
	OutputSetVolume(volume);
}

/*
 * get_output_time - return the number of milliseconds of
 * audio that have actually been played.
 */
int GetOutputTime()
{
	//TRACE(_T("GetOutputTime\n"));
	MMTIME time;

	if (!device)
		return 0;

	/*
	 * want time in bytes as this is the most accurate
	 */
	time.wType = TIME_BYTES;

	if (waveOutGetPosition(device, &time, sizeof(MMTIME)) == MMSYSERR_NOERROR)
	{
		/*
																	 * waveOutGetPosition apparently can return any type of time it
																	 * feels like if the one you specified wasn't good enough for 
																	 * it... sorry, "not supported".
																	 */
		switch (time.wType)
		{
		case TIME_BYTES:
			return pos + MulDiv(time.u.cb, 1000, bytes_per_sec);
		case TIME_MS:
			return pos + time.u.ms;
		case TIME_SAMPLES:
			return pos + MulDiv(time.u.sample, 1000, format.samplerate);
		default:
			break;
		}
	}

	/*
	 * worst case - return our own stored value of the time. this value is
	 * only accurate to the nearest block and is incremented in the callback
	 * function. using this with particularly large blocks (ie > 8k) will 
	 * cause the internal vis to become jumpy, ideally we want the above
	 * to work.
	 */
	return pos + MulDiv(output, 1000, bytes_per_sec);
}

/*
 * get_written_time - return the number of milliseconds of
 * audio that have been written to this plug-in but not necessarily
 * played (could be still buffered).
 */
int GetWrittenTime()
{
	//TRACE(_T("GetWrittenTime\n"));

	if (!device)
		return 0;

	return pos + MulDiv(written, 1000, bytes_per_sec);
}

/*
 * waveOut callback - don't call at all, ever. this is called by 
 * the windows waveOut subsystem thingy. and guess what - you really
 * can't call waveOut functions from this without deadlocking the
 * system (at some point). accurate MS documentation...
 *
 * it would be brilliant to be able to call waveOutUnprepareHeader here
 * since all headers passed to this have been played and are thus prepared.
 * the fact that you can't effectively renders using a callback function 
 * useless (well, almost).
 */
static void CALLBACK waveout_callback(HWAVEOUT device, UINT msg, DWORD inst, DWORD parm1, DWORD parm2)
{
	WAVEHDR* current = (WAVEHDR*)parm1;

	if (msg == WOM_DONE)
	{
		assert(current);

		/*
														 * we've now played current->dwBufferLength more bytes
														 */
		output += current->dwBufferLength;
		/*
														 * update the number of free headers
														 */
		EnterCriticalSection(&critical_section);
		free_headers++;
		LeaveCriticalSection(&critical_section);
	}
}
/*
 * get_statistics - get some stats on the current playback
 */
void get_statistics(stats_t* stats)
{
	assert(stats);

	stats->active = IsPlaying();
	stats->block_count = config_current.block_count;
	stats->block_usage = config_current.block_count - free_headers;
	stats->bytes_played = output;
	stats->bytes_written = written;
	stats->total_buffer_size = config_current.block_count * config_current.block_size;
	stats->total_buffer_usage = stats->total_buffer_size - CanWrite();
	stats->latency = GetWrittenTime() - GetOutputTime();
	stats->start_position = pos;
	stats->block_size = config_current.block_size;
	stats->format = format;
}

/*
 * get the master configuration
 */
void get_config(config_t* config)
{
	assert(config);
	*config = config_master;
}

/*
 * set the master configuration
 */
void set_config(const config_t* config)
{
	assert(config);
	config_master = *config;
}
