#include "stdafx.h"
#include "resource.h"   	// main symbols
#include "PlaylistManager.h"
#include "SysTrayPlay.h"
#include "MainHelper.h"
#include "MainWnd.h"
#include "PlaylistDlg.h"
#include "In/MultiDllInput.h"
#include "Out/MultiDllOutput.h"
#include "WAEmuWnd.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

HANDLE hTagsThread = NULL;
HANDLE hAutomationThread = NULL;
extern CStringArray genres;

DWORD q;

BOOL operator >(CachedTag& first, CachedTag& second)
{
	CString firstPath = first.GetDir();
	CString secondPath = second.GetDir();
	CString firstFilename = first.GetFilename();
	CString secondFilename = second.GetFilename();
	if (firstPath.CompareNoCase(secondPath) != 0)
		return (firstPath.CompareNoCase(secondPath) > 0);
	return (firstFilename.CompareNoCase(secondFilename) > 0);
}

BOOL operator <(CachedTag& first, CachedTag& second)
{
	CString firstPath = first.GetDir();
	CString secondPath = second.GetDir();
	CString firstFilename = first.GetFilename();
	CString secondFilename = second.GetFilename();
	if (firstPath.CompareNoCase(secondPath) != 0)
		return (firstPath.CompareNoCase(secondPath) < 0);
	return (firstFilename.CompareNoCase(secondFilename) < 0);
}


CachedTag::CachedTag()
{
	duration = 0;
	genre = 0;
	read = false;
	artist = "";
	album = "";
	comment = "";
	composer = "";
	playlist = "";
	title = "";
	year = "";
	TCHAR path[MAX_PATH];
	strcpy(path, (TCHAR *)(LPCTSTR)playlist);
	_splitpath(path, drive, dir, fname, ext);
}

CachedTag::CachedTag(const CachedTag& copie)
{
	read = copie.read;
	duration = copie.duration;
	genre = copie.genre;
	artist = copie.artist;
	album = copie.album;
	comment = copie.comment;
	composer = copie.composer;
	playlist = copie.playlist;
	title = copie.title;
	year = copie.year;
	TCHAR path[MAX_PATH];
	strcpy(path, (TCHAR *)(LPCTSTR)playlist);
	_splitpath(path, drive, dir, fname, ext);
}

CachedTag::CachedTag(CString nPlaylist)
{
	read = false;
	duration = 0;
	genre = 0;
	artist = "";
	album = "";
	comment = "";
	composer = "";
	playlist = nPlaylist;
	title = "";
	year = "";
	TCHAR path[MAX_PATH];
	strcpy(path, (TCHAR *)(LPCTSTR)playlist);
	_splitpath(path, drive, dir, fname, ext);
}

CachedTag::~CachedTag()
{
}

void CachedTag::copy(CachedTag copie)
{
	read = copie.read;
	duration = copie.duration;
	genre = copie.genre;
	artist = copie.artist;
	album = copie.album;
	comment = copie.comment;
	composer = copie.composer;
	playlist = copie.playlist;
	title = copie.title;
	year = copie.year;
	TCHAR path[MAX_PATH];
	strcpy(path, (TCHAR *)(LPCTSTR)playlist);
	_splitpath(path, drive, dir, fname, ext);
}

CString CachedTag::GetAlbum()
{
	if (!read)
		ReadTag();
	return album;
}

CString CachedTag::GetArtist()
{
	if (!read)
		ReadTag();
	return artist;
}

CString CachedTag::GetComment()
{
	if (!read)
		ReadTag();
	return comment;
}

CString CachedTag::GetComposer()
{
	if (!read)
		ReadTag();
	return composer;
}

int CachedTag::GetGenreI()
{
	if (!read)
		ReadTag();
	return genre;
}


CString CachedTag::GetGenre()
{
	if (!read)
		ReadTag();
	return genres[GetGenreI()];
}

CString CachedTag::GetFilenameExt()
{
	CString filename = playlist;
	TCHAR path[MAX_PATH];
	TCHAR drive[MAX_PATH];
	TCHAR dir[MAX_PATH];
	TCHAR fname[MAX_PATH];
	TCHAR ext[MAX_PATH];
	strcpy(path, (TCHAR *)(LPCTSTR)playlist);
	_splitpath(path, drive, dir, fname, ext);
	filename.Format(_T("%s%s"), fname, ext);
	return filename;
}

CString CachedTag::GetDuration()
{
	if (!read)
		ReadTag();
	CString dur = "mm:ss";
	if (duration != 0)
		dur.Format(_T(" %02d:%02d"), (int)(duration / 60), duration % 60);
	return dur;
}

int CachedTag::GetDurationI()
{
	return duration;
}

CString CachedTag::GetPlaylist()
{
	return playlist;
}

bool CachedTag::GetRead()
{
	return read;
}

CString CachedTag::GetTag(int ShowingMode)
{
	CString CurrentSong = "";

	if (ShowingMode == 2)
		CurrentSong = GetFilenameExt();
	else if (ShowingMode == 1)
	{
		if (GetArtist().GetLength() == 0)
			CurrentSong = GetTitle();
		else
			CurrentSong = GetArtist() + " - " + GetTitle();
	}
	else if (ShowingMode == 0)
	{
		char info[260];
		int length = 0;
		stdIn->GetFileInfo((TCHAR *)(LPCTSTR)GetPlaylist(), info, &length);
		CurrentSong.Format(_T("%s"), info);
	}
	return CurrentSong;
}

CString CachedTag::GetTitle()
{
	if (!read)
		ReadTag();
	return title;
}

CString CachedTag::GetYear()
{
	if (!read)
		ReadTag();
	return year;
}

CString CachedTag::GetDrive()
{
	CString tmp;
	tmp.Format(_T("%s"), drive);
	return tmp;
}

CString CachedTag::GetDir()
{
	CString tmp;
	tmp.Format(_T("%s"), dir);
	return tmp;
}

CString CachedTag::GetFilename()
{
	CString tmp;
	tmp.Format(_T("%s"), fname);
	return tmp;
}

CString CachedTag::GetExt()
{
	CString tmp;
	tmp.Format(_T("%s"), ext);
	return tmp;
}

void CachedTag::ReadTag(bool forceRead)
{
	if (read && !forceRead)
		return;
	char filename[MAX_PATH];
	char meta[10];
	char ret[MAX_PATH];

	strcpy(filename, (LPCSTR)GetPlaylist());

	extendedFileInfoStruct eFIStruct;

	eFIStruct.filename = (char*)&filename;
	eFIStruct.metadata = (char*)&meta;
	eFIStruct.ret = (char*)&ret;

	stdIn->GetFileInfo(filename, (TCHAR *)(LPCTSTR)ret, &eFIStruct.retlen);
	if(eFIStruct.retlen!=0)
		SetTitle(eFIStruct.ret);
	else
		SetTitle(GetFilename());
	SetDuration(eFIStruct.retlen / 1000);

	eFIStruct.retlen = MAX_PATH;

	strcpy(meta, "length");
	memset(eFIStruct.ret, 0, MAX_PATH);
	stdIn->winampGetExtendedFileInfo(eFIStruct);
	if (strlen(eFIStruct.ret) > 0)
	{
		SetDuration(atoi(eFIStruct.ret) / 1000);
	}
	eFIStruct.retlen = MAX_PATH;
	strcpy(meta, "title");
	memset(eFIStruct.ret, 0, MAX_PATH);
	stdIn->winampGetExtendedFileInfo(eFIStruct);
	if (strlen(eFIStruct.ret) > 0)
	{
		SetTitle(eFIStruct.ret);
	}
	eFIStruct.retlen = MAX_PATH;
	strcpy(meta, "artist");
	memset(eFIStruct.ret, 0, MAX_PATH);
	stdIn->winampGetExtendedFileInfo(eFIStruct);
	if (strlen(eFIStruct.ret) > 0)
	{
		SetArtist(eFIStruct.ret);
	}
	eFIStruct.retlen = MAX_PATH;
	strcpy(meta, "comment");
	memset(eFIStruct.ret, 0, MAX_PATH);
	stdIn->winampGetExtendedFileInfo(eFIStruct);
	if (strlen(eFIStruct.ret) > 0)
	{
		SetComment(eFIStruct.ret);
	}
	eFIStruct.retlen = MAX_PATH;
	strcpy(meta, "album");
	memset(eFIStruct.ret, 0, MAX_PATH);
	stdIn->winampGetExtendedFileInfo(eFIStruct);
	if (strlen(eFIStruct.ret) > 0)
	{
		SetAlbum(eFIStruct.ret);
	}
	eFIStruct.retlen = MAX_PATH;
	strcpy(meta, "year");
	memset(eFIStruct.ret, 0, MAX_PATH);
	stdIn->winampGetExtendedFileInfo(eFIStruct);
	if (strlen(eFIStruct.ret) > 0)
	{
		SetYear(eFIStruct.ret);
	}
	eFIStruct.retlen = MAX_PATH;
	strcpy(meta, "genre");
	memset(eFIStruct.ret, 0, MAX_PATH);
	stdIn->winampGetExtendedFileInfo(eFIStruct);
	if (strlen(eFIStruct.ret) > 0)
	{
		SetGenreI(atoi(eFIStruct.ret));
	}
}
void CachedTag::SetAlbum(CString nAlbum)
{
	read = true;
	album = nAlbum;
}

void CachedTag::SetArtist(CString nArtist)
{
	read = true;
	artist = nArtist;
}

void CachedTag::SetComment(CString nComment)
{
	read = true;
	comment = nComment;
}

void CachedTag::SetComposer(CString nComposer)
{
	read = true;
	composer = nComposer;
}

void CachedTag::SetDuration(int nDuration)
{
	read = true;
	duration = nDuration;
}

void CachedTag::SetGenreI(int nGenre)
{
	read = true;
	genre = nGenre;
}

void CachedTag::SetPlaylist(CString nplaylist)
{
	playlist = nplaylist;
	TCHAR path[MAX_PATH];
	strcpy(path, (TCHAR *)(LPCTSTR)playlist);
	_splitpath(path, drive, dir, fname, ext);
}

void CachedTag::SetTitle(CString nTitle)
{
	read = true;
	title = nTitle;
}

void CachedTag::SetYear(CString nYear)
{
	read = true;
	year = nYear;
}


PlaylistManager::PlaylistManager()
{
	maxSize = 0;
	taille = 0;
#ifndef QCACHED
	CachedTags = (CachedTag * *)calloc(maxSize, sizeof(CachedTag *));
#else
	CachedTags.RemoveAll();
#endif
}

PlaylistManager::~PlaylistManager()
{
#ifndef QCACHED
	RemoveAll();
	delete CachedTags;
#else
	CachedTags.RemoveAll();
#endif
}

void PlaylistManager::Add(CString tag, int index)
{
	if (taille == MAXFILE)
		return;
	try
	{
		CachedTag tmp(tag);
		if (index == -1)
			InsertAt(tmp, taille);
		else
		{
#ifndef QCACHED
			(*(CachedTags + index))->copy(tmp);
#else
			CachedTags[index].copy(tmp);
#endif
		}
	}
	catch (...)
	{
	}
}

void PlaylistManager::Add(CachedTag tag, int index)
{
	if (taille == MAXFILE)
		return;
	try
	{
		if (index == -1)
			InsertAt(tag, taille);
		else
		{
#ifndef QCACHED
			(*(CachedTags + index))->copy(tag);
#else
			CachedTags[index].copy(tag);
#endif
		}
	}
	catch (...)
	{
	}
}

void PlaylistManager::InsertAt(CachedTag tag, int index)
{
	if (taille == MAXFILE)
		return;
	try
	{
		if (++taille == maxSize + 1)
		{
			maxSize = maxSize + PAS;
#ifndef QCACHED
			CachedTags = (CachedTag * *)realloc(CachedTags, maxSize * sizeof(CachedTag *));
			for (int i = maxSize - PAS; i < maxSize; i++)
				*(CachedTags + i) = new CachedTag();
#endif
		}
#ifndef QCACHED
		CachedTag* tmp = *(CachedTags + taille - 1);
		memmove(CachedTags + index + 1, CachedTags + index, (taille - index - 1) * sizeof(CachedTag *));
		*(CachedTags + index) = tmp;
		(*(CachedTags + index))->copy(tag);
#else
		CachedTag tmp(tag);
		CachedTags.InsertAt(index, tmp);
#endif
	}
	catch (...)
	{
	}
}

CachedTag PlaylistManager::Get(int index)
{
	try
	{
		CachedTag tmp;
		if (index < taille)
		{
#ifndef QCACHED
			return **(CachedTags + index);
#else
			return CachedTags[index];
#endif
		}
		else
			return tmp;
	}
	catch (...)
	{
		CachedTag tmp;
		tmp.SetAlbum("");
		tmp.SetArtist("");
		tmp.SetComment("");
		tmp.SetComposer("");
		tmp.SetDuration(0);
		tmp.SetGenreI(0);
		tmp.SetPlaylist("");
		tmp.SetTitle("");
		tmp.SetYear("");
		return tmp;
	}
}

int PlaylistManager::GetIndex(CString filename)
{
	int posIndex;
	//TRACE(_T("GetIndex %s %d\n"), filename, GetSize());
	if (GetSize() == 0)
		return -1;
	for (posIndex = 0; posIndex < GetSize() && Get(posIndex).GetPlaylist().Compare(filename); posIndex++)
		;
	if (posIndex == GetSize())
		return -1;
	return posIndex;
}

CString PlaylistManager::GetAlbum(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetAlbum();
#else
			return CachedTags[index].GetAlbum();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

CString PlaylistManager::GetArtist(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetArtist();
#else
			return CachedTags[index].GetArtist();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

CString PlaylistManager::GetComment(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetComment();
#else
			return CachedTags[index].GetComment();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

CString PlaylistManager::GetComposer(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetComposer();
#else
			return CachedTags[index].GetComposer();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

CString PlaylistManager::GetDuration(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetDuration();
#else
			return CachedTags[index].GetDuration();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

void PlaylistManager::SetDuration(int index, int nDuration)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			(*(CachedTags + index))->SetDuration(nDuration);
#else
			CachedTags[index].SetDuration(nDuration);
#endif
		}
	}
	catch (...)
	{
	}
}
int PlaylistManager::GetDurationI(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetDurationI();
#else
			return CachedTags[index].GetDurationI();
#endif
		}
		else
			return 0;
	}
	catch (...)
	{
		return 0;
	}
}

CString PlaylistManager::GetFilenameExt(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetFilenameExt();
#else
			return CachedTags[index].GetFilenameExt();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

CString PlaylistManager::GetGenre(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetGenre();
#else
			return CachedTags[index].GetGenre();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

int PlaylistManager::GetGenreI(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetGenreI();
#else
			return CachedTags[index].GetGenreI();
#endif
		}
		else
			return 0;
	}
	catch (...)
	{
		return 0;
	}
}

CString PlaylistManager::GetPlaylist(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetPlaylist();
#else
			return CachedTags[index].GetPlaylist();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

bool PlaylistManager::GetRead(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetRead();
#else
			return CachedTags[index].GetRead();
#endif
		}
		return false;
	}
	catch (...)
	{
		return false;
	}
}
int PlaylistManager::GetShowingMode()
{
	return ShowingMode;
}

CString PlaylistManager::GetTag(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetTag(ShowingMode);
#else
			return CachedTags[index].GetTag(ShowingMode);
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

CString PlaylistManager::GetTitle(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetTitle();
#else
			return CachedTags[index].GetTitle();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

void PlaylistManager::SetTitle(int index, CString title)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			(*(CachedTags + index))->SetTitle(title);
#else
			CachedTags[index].SetTitle(title);
#endif
		}
	}
	catch (...)
	{
	}
}

CString PlaylistManager::GetYear(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetYear();
#else
			return CachedTags[index].GetYear();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

CString PlaylistManager::GetDrive(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetDrive();
#else
			return CachedTags[index].GetDrive();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

CString PlaylistManager::GetDir(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetDir();
#else
			return CachedTags[index].GetDir();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

CString PlaylistManager::GetFilename(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetFilename();
#else
			return CachedTags[index].GetFilename();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

CString PlaylistManager::GetExt(int index)
{
	try
	{
		if (index < taille)
		{
#ifndef QCACHED
			return (*(CachedTags + index))->GetExt();
#else
			return CachedTags[index].GetExt();
#endif
		}
		else
			return "";
	}
	catch (...)
	{
		return "";
	}
}

int PlaylistManager::GetSize()
{
	return taille;
}

void PlaylistManager::SetShowingMode(int nShowingMode)
{
	ShowingMode = nShowingMode;
}

bool PlaylistManager::StopReadAll()
{
	bool toret = (hTagsThread != NULL);
	if (toret)
	{
		::TerminateThread(hTagsThread, 0);
		hTagsThread = NULL;
	}
	return toret;
}

bool PlaylistManager::StopPerformAutomation()
{
	bool toret = (hAutomationThread != NULL);
	if (toret)
	{
		::TerminateThread(hAutomationThread, 0);
		hAutomationThread = NULL;
	}
	return toret;
}

void ReadAllTagsThread()
{
	try
	{
		for (UINT i = 0; i < (UINT)m_PlaylistManager.GetSize(); i++)
		{
			m_PlaylistManager.ReadTagAT(i, true);
			/*		if (PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
																																			{
																																				PlaylistDlg->ShowTimes();
																																				PlaylistDlg->m_playlistat.RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW);
																																			}
																																	*/
			Sleep(77);
		}
		if (PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
		{
			PlaylistDlg->ShowTimes();
			PlaylistDlg->m_playlistat.RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW);
		}
	}
	catch (...)
	{
	}
	hTagsThread = NULL;
}

void PlaylistManager::ReadAllTags()
{
	if (hTagsThread == NULL)
	{
		hTagsThread = ::CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ReadAllTagsThread, NULL, 0, &q);
		::SetThreadPriority(hTagsThread, THREAD_PRIORITY_LOWEST);
	}
}
void PerformAutomationThread()
{
	bool b = false;
	if (pMainFrame->AutRemove || pMainFrame->AutRemovedup || pMainFrame->AutRescan || pMainFrame->AutSort)
		b = m_PlaylistManager.StopReadAll();

	if (PlaylistDlg != NULL)
	{
		PlaylistDlg->StopRemoveDuplicateEntries();
		PlaylistDlg->StopRemovene();
		PlaylistDlg->StopSort();
	}

	if (pMainFrame->AutRescan)
		b = true;

	if (pMainFrame->AutRemove)
	{
		UINT i = 0;
		bool CurDeleted = false, deleted = false;
		while (i < (UINT)m_PlaylistManager.GetSize() && (UINT)m_PlaylistManager.GetSize() > 0)
		{
			if (!pMainFrame->FileExists(m_PlaylistManager.GetPlaylist(i)))
			{
				deleted = true;
				if (PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
					PlaylistDlg->m_playlistat.DeleteString(i);
				m_PlaylistManager.RemoveAt(i);
				if (i < pMainFrame->Current)
					pMainFrame->Current--;
				else if (i == pMainFrame->Current && !CurDeleted)
					CurDeleted = true;
			}
			else
				i++;
			Sleep(1);
		}
		if (CurDeleted)
		{
			pMainFrame->Current = 0;
			if (m_PlaylistManager.GetSize() > 0)
			{
				m_PlaylistManager.ReadTagAT(pMainFrame->Current);
				if (::SendMessage(WAEmuWnd->m_hWnd, WM_WA_IPC, 0, IPC_ISPLAYING) != PLAYER_MODE_CLOSE)
					::SendMessage(WAEmuWnd->m_hWnd, WM_COMMAND, WINAMP_BUTTON4, 0);
			}
			else
			{
				if (::SendMessage(WAEmuWnd->m_hWnd, WM_WA_IPC, 0, IPC_ISPLAYING) != PLAYER_MODE_CLOSE)
					::SendMessage(WAEmuWnd->m_hWnd, WM_COMMAND, WINAMP_BUTTON4, 0);
			}
			pMainFrame->UpdMBTR();
		}
		if (PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL && deleted)
			PlaylistDlg->m_playlistat.SetSel(-1, FALSE);

		if (PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
		{
			PlaylistDlg->ShowFileName();
			PlaylistDlg->m_playlistat.Invalidate();
		}
	}

	if (pMainFrame->AutSort || pMainFrame->AutRemovedup)
	{
		m_PlaylistManager.QuickSort(TRUE);

		if (PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
		{
			PlaylistDlg->ShowFileName();
			PlaylistDlg->m_playlistat.SetTopIndex(PlaylistDlg->m_playlistat.GetFirstSelected() != -1 ? PlaylistDlg->m_playlistat.GetFirstSelected() :
												  pMainFrame->Current);
			PlaylistDlg->m_playlistat.Invalidate();
		}
	}

	if (pMainFrame->AutRemovedup)
	{
		UINT i = 0, j;
		bool deleted = false;
		while (i < (UINT)m_PlaylistManager.GetSize() && (UINT)m_PlaylistManager.GetSize() > 0)
		{
			j = i + 1;
			while (j < (UINT)m_PlaylistManager.GetSize())
			{
				if (m_PlaylistManager.GetPlaylist(i).Compare(m_PlaylistManager.GetPlaylist(j)) == 0)
				{
					if (PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
						PlaylistDlg->m_playlistat.DeleteString(j);
					deleted = true;
					m_PlaylistManager.RemoveAt(j);
					if (j < pMainFrame->Current)
						pMainFrame->Current--;
					else if (j == pMainFrame->Current)
						pMainFrame->Current = i;
				}
				else
					j++;
				Sleep(1);
			}
			i++;
		}
		if (PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL && deleted)
			PlaylistDlg->m_playlistat.SetSel(-1, FALSE);

		if (PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
		{
			PlaylistDlg->ShowFileName();
			PlaylistDlg->m_playlistat.Invalidate();
		}
	}


	if (pMainFrame->AutRescan)
	{
		UINT PLLPrev = m_PlaylistManager.GetSize();
		for (int i = 0; i < pMainFrame->AutDirs.GetSize(); i++)
			pMainFrame->FindMP3Files(pMainFrame->AutDirs[i], true);

		m_PlaylistManager.QuickSort(TRUE);

		if ((pMainFrame->RandomFirst || pMainFrame->ShuffleMode) && PLLPrev == 0)
			pMainFrame->FirstInMyRandom = pMainFrame->Current = pMainFrame->MyRandom(true, true);
		else if (PLLPrev == 0)
		{
			pMainFrame->Current = 0;pMainFrame->FirstInMyRandom = 0xFFFFFFFF;
		}

		if (PlaylistDlg != NULL && PlaylistDlg->m_hWnd != NULL)
		{
			if (m_PlaylistManager.GetSize() != PLLPrev)
			{
				PlaylistDlg->FillPlaylist();
				PlaylistDlg->ShowFileName();
			}
			PlaylistDlg->m_playlistat.Invalidate();
		}

		if (PLLPrev == 0)
		{
			if (::SendMessage(WAEmuWnd->m_hWnd, WM_WA_IPC, 0, IPC_ISPLAYING) != PLAYER_MODE_CLOSE)
				::SendMessage(WAEmuWnd->m_hWnd, WM_COMMAND, WINAMP_BUTTON4, 0);

			bool first = true;
			int q = 0;
			do
			{
				if (!first)
				{
					if (pMainFrame->Current < (UINT)m_PlaylistManager.GetSize() - 1)
						pMainFrame->Current++;
					else
					{
						pMainFrame->Current = 0;
						q++;
						if (pMainFrame->m_Repeat != 1)
							break;
					}
				}
				else
					first = false;
				break;
				Sleep(77);
			}
			while (q < 2);
			m_PlaylistManager.ReadTagAT(pMainFrame->Current);
			pMainFrame->UpdMBTR();
		}
	}

	if (b)
		m_PlaylistManager.ReadAllTags();
}

void PlaylistManager::PerformAutomation()
{
	if (PlaylistDlg != NULL)
	{
		if (PlaylistDlg->RemoveDuplicateEntriesThreadHan != NULL || PlaylistDlg->RemoveneThreadHan != NULL || PlaylistDlg->SortThreadHan != NULL)
			return;
	}

	if (hAutomationThread == NULL)
	{
		DWORD q;
		hAutomationThread = ::CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)PerformAutomationThread, NULL, 0, &q);
		::SetThreadPriority(hAutomationThread, THREAD_PRIORITY_LOWEST);
	}
}
void PlaylistManager::ReadTagAT(int index, bool forceRead)
{
	if (index < taille)
	{
#ifndef QCACHED
		(*(CachedTags + index))->ReadTag(forceRead);
#else
		CachedTags[index].ReadTag(forceRead);
#endif
	}
}
void PlaylistManager::RemoveAt(int index)
{
#ifndef QCACHED
	CachedTag* tmp = *(CachedTags + index);
	memmove(CachedTags + index, CachedTags + index + 1, (taille - index - 1) * sizeof(CachedTag *));
	*(CachedTags + taille - 1) = tmp;
#else
	CachedTags.RemoveAt(index);
#endif
	taille--;
	if (taille + PAS == maxSize)
	{
#ifndef QCACHED
		for (int i = maxSize - PAS; i < maxSize; i++)
			delete * (CachedTags + i);
#endif
		maxSize = maxSize - PAS;
#ifndef QCACHED
		CachedTags = (CachedTag * *)realloc(CachedTags, maxSize * sizeof(CachedTag *));
#endif
	}
}

void PlaylistManager::RemoveAll()
{
#ifndef QCACHED
	for (int i = taille - 1; i >= 0; i--)
		RemoveAt(i);
#else
	CachedTags.RemoveAll();
	taille = 0;
	maxSize = 0;
#endif
}

void PlaylistManager::QuickSort(BOOLEAN asc)
{
#ifndef QCACHED
	pMainFrame->SortRec(0, m_PlaylistManager.GetSize() - 1);
#else
	CString playlist = "";
	if (pMainFrame->Current >= 0 && pMainFrame->Current < CachedTags.GetSize())
		playlist = GetPlaylist(pMainFrame->Current);
	CachedTags.QuickSort(asc);
	if (pMainFrame->Current >= 0 && pMainFrame->Current < CachedTags.GetSize())
		pMainFrame->Current = GetIndex(playlist);
#endif
}
